/* owl.c
 * Copyright 2005-2009 Antonio Maschio <tbin at libero dot it>
 *
 * owl - Obfuscated Weird Language
 * 
 * (source code optimized for 80 columns with 4 chars tab stops)
 * (source code edited and maintained with gvim)
 *
 * termios statement:
 * termios features for command '(' are taken from file term2.c 
 * under the following copyright:
 * Copyright (C) 2000 Kyle R. Burton, All Rights Reserved
 * mortis@voicenet.com
 * http://www.voicenet.com/~mortis
 * 
 * owl statement:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * It's GPL! Enjoy! */

#include "owl.h"

/****************************
 *        FUNCTIONS         *
 ****************************/

/* printHelp()
 * help on owl usage; execute -h/--help options */
void printHelp(void) {
    printf("Usage: owl [options] [file] [commands]\n");
    printf("Options:\n");
#ifdef LEAVESEXE
    printf("  -c               compile source an leave an executable\n");
#endif
    printf("      --conditions print redistribution conditions from GPL3\n");
#ifdef OWL_DEBUG
    printf("  -d               execute program in debug mode\n");
    printf("  -D               if in debug mode, print var contents\n");
#endif
    printf("  -e               toggle PAD's erasure before string initialization\n");
    printf("  -h, --help       display this information\n");
    printf("  -i               toggle integer division mode\n");
    printf("  -p \'<string>\'    Parse <string> enclosed in quotes and exit\n");
    printf("  -r               toggle rounding mode\n");
#ifdef LEAVESASM
    printf("  -s               compile source and leave an assembly file\n");
#endif
#ifndef OWL_REMOVE
    printf("  -S               if in debug mode, enable single-step running\n");
    printf("  -t               print a timer report after execution\n");
#endif
    printf("  -v, --version    display version, license and warranty\n");
    printf("      --warranty   print warranty conditions from GPL3\n");
    printf("\nCode contained into <commands> will be executed before program start,\n");
    printf("but not if -p option is used.\n");
    printf("If no <file> is specified or no <string> is given to option -p\n");
    printf("owl will complain.\n");

    printf("\nFor bug reporting, please write to <tbin at libero dot it>.\n\n"); 
}


/* printVersion()
 * version message; execute -v/--version options */
void printVersion(int flag) {
    printf("\nowl (Obfuscated Weird Language) interpreter\n");
    printf("version %d.%d.%d (build %d, %s, %s)\n", MAJOR, MINOR, RELEASE, BUILD, __DATE__, __TIME__);
    printf("Copyright (C) %s %s <%s>. ", CY, AU, EM);
    if (flag) printf("This program\ncomes with ABSOLUTELY NO WARRANTY; for details type 'owl --warranty'.\n");
    if (flag) printf("This is free software, and you are welcome to redistribute it under certain\n");
    if (flag) printf("conditions; type 'owl --conditions' for details.");
    printf("\nThis software is released under the terms of the GNU General Public License 3.\n\n");
}


/* printWarranty()
 * warranty message (required by GPL3); execute --warranty option */
void printWarranty(int flag) {
    if (flag) printVersion(0);
    printf("THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY\n");
    printf("APPLICABLE LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT\n");
    printf("HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM \"AS IS\" WITHOUT\n");
    printf("WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT\n");
    printf("LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A\n");
    printf("PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF\n");
    printf("THE PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME\n");
    printf("THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.\n\n");
}


/* printConditions()
 * conditions message (required by GPL3); execute --conditions option */
void printConditions(int flag) {
    if (flag) printVersion(0);
    printf("You may make, run and propagate covered works that you do not convey,\n");
    printf("without conditions so long as your license otherwise remains in force.\n");
    printf("You may convey covered works to others for the sole purpose of having\n");
    printf("them make modifications exclusively for you, or provide you with\n");
    printf("facilities for running those works, provided that you comply with the\n");
    printf("terms of this License in conveying all material for which you do not\n");
    printf("control copyright. Those thus making or running the covered works for\n");
    printf("you must do so exclusively on your behalf, under your direction and\n");
    printf("control, on terms that prohibit them from making any copies of your\n");
    printf("copyrighted material outside their relationship with you.\n\n");
}


/* erase_buffer()
 * erase buffer up to LIMIT 
 * WARNING: only to be used for owl char buffers (sized with 'LIMIT') */
void erase_buffer(char *buffer) {
    int i;
    for (i = 0; i < LIMIT; i++) *buffer++ = '\0';
}


/* pop()
 * pop integer */
long long int pop(void) {
    if (TOS <= 0) {
        fprintf(stderr,UNDERFLOWERROR);
        quit();
    }
    else return Stack[TOS--];
}


/* push()
 * push integer */
void push(long long int n) {
    if (TOS > (STACKLIMIT - 1)) {
        fprintf(stderr, OVERFLOWERROR);
        quit();
    }
    else Stack[++TOS] = n;
}


/* getVar()
 * subroutine for variables: if a letter is followed by a special operator, 
 * then execute relative function, otherwise push letter's ASCII value;
 * some parts of dynamic binding are treated here; 
 * return updated position in the input stream */
char* getVar(char *input) {
    int vtype = *input;   // var type: a-z=functions, A-Z=numbers 
    int done  = 0;        // offset from current input stream position 

    switch(*(input + 1)) {
        case '_':
            /* a_, dynamic binding  
             * if '_' is followed by ',' store PAD into function as commands */
            if ((',' == *(input + 2)) && islow(vtype)) {
                strncpy((VarFBuffer[vtype - 'a']), PAD, LIMIT);
                done = 2;
            }
            /* a_' dynamic binding  
             * if '_' is followed by ' store PAD into function as string */
            else if (('\'' == *(input + 2)) && islow(vtype)) {
                char line[LIMIT];
                strcpy(line, "\"");
                strncat(line, PAD, LIMIT-3); // terminator and quotes
                strcat(line, "\"");
                strncpy((VarFBuffer[vtype - 'a']), line, LIMIT);
                done = 2;
            }
            /* a_ or A_
             * push ASCII value of character before underscore; in next 
             * iteration, the underscore will be treated in parse() */
            else push(vtype);
            break;
        case ',':
            /* ,, dynamic binding 
             * if ',' is followed by another ',' copies var into PAD*/
            if ((',' == *(input + 2)) && islow(vtype)) {
                strncpy(PAD, VarFBuffer[vtype - 'a'], LIMIT);
                done = 2;
            }
            /* A, store number */
            else if (iscap(vtype)) {
                VarIBuffer[vtype - 'A'] = pop();
                done = 1;
            }
            /* a, store function */
            else if (islow(vtype)) {
                strncpy((VarFBuffer[vtype - 'a']), FuncBuffer, LIMIT);
                erase_buffer(FuncBuffer);
                erase_buffer(BackFuncBuffer);
                done = 1;
            }
            break;    
        case '@':
            /* a@, dynamic binding 
             * if '@' is followed by ',' store function index */
            if ((',' == *(input + 2)) && islow(vtype)) {
                fIndex = vtype;
                done = 2;
            }
            /* A@ fetch number */
            else if (iscap(vtype)) {
                push(VarIBuffer[vtype - 'A']);
                done = 1;
            }
            /* a@ execute function */
            else if (islow(vtype)) {
                parse(VarFBuffer[vtype - 'a']);
                done = 1;
            }
            break;
        default:
            /* a or A  push ASCII value */
            push(vtype);
            break;
    }
    return (input + done);
}


/* quit()
 * quit program, emitting TOS or EXIT_FAILURE */
void quit(void) {
    fprintf(stdout, "\n");
    if (TOS) exit(pop());
    else exit(EXIT_FAILURE);
}


/* fprintb()
 * print an integer in binary format according to the following format:
 * - bit 0 appears on the right (e.g. 13d => 1101b)
 * - negative numbers occupy all 64 bits  */
void fprintb(FILE *file, long long int n) {
    int num[BITS], c = -1;
    if (!n) fprintf(file, "0");
    while (n && (c < (BITS - 1))) num[++c] = n & 1, n >>= 1;
    while (c >= 0) fprintf(file, "%d", num[c--]);
}


/* fileopen()
 * open a file and try adding extension .owl if not found;
 * return the file descriptor of the opened file, if found in either way,
 * or null if neither is found */
FILE* fileopen(char* original) {
    char fname[LIMIT], *fn=fname;
    FILE * ff = fopen(original, "r");

    if (!ff) {
        strncpy(fname,original,LIMIT-1);
        fn+=strlen(fname)-1;
        while (strlen(fname) > LIMIT-5) *fn--='\0';
        strcat(fname, ".owl");
        ff=fopen(fname, "r");
        if (!ff) return NULL;
    }
    return ff;
}


/* include_file()
 * load and execute file, retrieving its name from input stream,
 * enclosed by '_][' or '_[]'; the included file is executed as a standalone 
 * program, with current stack and current variables as environment;
 * return file name length (actually unused), or exit with EXIT_FAILURE if 
 * file is not found; the starting characters _[ or _] are detected into 
 * parse(), so they don't appear here */
int include_file(char *input) {
    char bufn[LIMIT], *p = bufn; 
    int ret = 0;
    FILE *f;    

    while (*(input) != '[' && *(input)!= ']') *p++ = *input++;
    *p = '\0';

    if (bufn[0]) {
        f = fileopen(bufn);
        if (!f) {
            fprintf(stderr, OPENFILEERROR, bufn);
            exit(EXIT_FAILURE);
        }
        else {
            ret = strlen(bufn);
            /* here bufn is reused for fgets */
            while (fgets(bufn, LIMIT, f) && !parse(bufn));
            fclose(f);
        }
    }
    return ret;
}


/* printvars()
 * print non null variables in the debug function */
void printvars(void) {
    int i;
    long long int s=0LL;
        /* is there any non null variable? */
        for(i=0; i<MAXVARS; i++) s += VarIBuffer[i];
        /* if so, print each with name and content */
        if (s) {
            fprintf(stdout, " -- ");
            for(i=0; i<MAXVARS; i++)
                if(VarIBuffer[i])
                    fprintf(stdout, "%c=%lld, ", 'A'+i, VarIBuffer[i]);
        }
}


/* divide()
 * calculate division according either to the classic mode or to the
 * Number Theory mode, from long long int arguments giving 
 * long long int result */
long long int divide(long long int dividend, long long int divisor) {
    long long int res;      // temporary integer result
    short k;                // k: add/subtract 1 to res (NT mode)
    long double resd;       // double result to be rounded 

    /* predetermined results for special cases: the dividend is returned 
     * in case of null divisor */
    if (!divisor) return dividend;
    /* calculate division: */
    resd = (long double)dividend/(long double)divisor;
    res = (long long int)resd;
    k = dividend - divisor * res != 0 ? 1 : 0;
    /* integer division according to Number Theory */
    if (isNTDiv) return (dividend<0)?((res<=0)?res-k:res+k):_round(resd);
    /* classic integer division according to _round */
    else return _round(resd);
}


/* nroot()
 * calculate nth root using an iterative algorithm 
 * from long long int parameters giving long long int result 
 * (see http://gmplib.org/manual/Nth-Root-Algorithm.html)
 * root is the greatest integer whose nth power is not larger than base */
long long int nroot(long long int base, long long int root) {
    long double a=0.0L, a1=1.0L;
    long double dbase, droot;
    short sign=1;
                
    /* predetermined results for special cases; 0 is a safe result for 
     * impossible roots; see J.W. Crenshaw, "Math tool-kit for Real Time 
     * Programming", CMP books, 2000, page 62, listing 4.3 (for square root) */
    if (root == 1LL) return base;
    else if (root < 0LL) return 0LL; // always zero in integer
    else if (base < 0LL && !(root % 2LL)) return 0LL; // impossible root
    
    /* calculate nth positive root of a positive base (sign retained) */
    if (base < 0LL) sign = -1, base = -base;
    dbase=(long double)base, droot=(long double)root;
    while (fabsl(a-a1) > EPSILON) {
        a=a1;
        a1=(dbase/powl(a,droot-1.0L)+(droot-1.0L)*a)/droot;
    }
    /* note: we don't round here, since a rounding toward the nearer integer
     * for decimals >= 0.5 would return a square larger than original base 
     * e.g. 981 3: is 9.9363, thus 10 rounded, whose cube is 1000 */ 
    return (long long int)(a1*sign);
}

 
/* power()
 * calculate power from long long int parameters, 
 * giving long long int result */
long long int power(long long int base, long long int exp) {
    long long int res = 1LL;

    /* predetermined results for special cases */
    if (exp == 0LL) return res;
    else if (exp == -1LL) return _round(1.0/base);
    else if (exp == 1LL || base == 1LL) return base;
    /* calculate nth power */
    else if (exp > 0LL) while (exp--) res*=base;
    else return 0LL;
    return res;
}


/* The DEBUGGER; for each command execution 
 * print stack content enclosed into square brackets and 
 * next character (instruction) to be executed; 
 * a debug line is independent of the program output
 * if ftell(stdout) returns other than -1 */
void debug(char *input) {
    int u;
    char c = *input, c1 = *(input+1);
    /* debug: skip control characters and sharp comments */
    if (c > ' ' && (c!='#' || (c=='#' && c1==',') || (c=='#' && c1=='@'))) {
#ifndef OWL_REMOVE
        /* debug: single step execution in debug mode */
        if (singleStep) {
            // thanks to Kyle R. Burton for this routine. 
            // See initial GPL banner for his credits
            struct termios modes, savemodes;
            int opg;
            tcgetattr(0, &modes);
            savemodes = modes;
            modes.c_lflag &= ~ICANON;
            modes.c_lflag &= ~ECHO;
            tcsetattr(0, 0, &modes);
            opg = fgetc(stdin);
            if (opg == 'S' || opg == 's') singleStep=FALSE;
            tcsetattr(0, 0, &savemodes);
        }
#endif
        dpos=ftell(stdout);
        /* debug: if printing position has changed, print CR.
         * Note: this feature requires ftell(stdout) to return a
         * positive long for the terminal in use; on some terminals
         * (real or emulated) the returned value is always -1, 
         * because we cannot seek the stream. In this case, the
         * fputc('\n, stdout) is never called, and the debug line 
         * appears after the last program output in the same line. 
         * At present, ftell(stdout) happens to work for Mac OS X 10.4
         * but not for Linux. */
        if ((dpos-dposex)>0 && !singleStep) fputc('\n', stdout);
            /* debug: skip 'pascal' comments (*..*) */
            if (!((c == '(') && (c1 == '*'))) {
                fprintf(stdout, "   [ ");
                /* debug: print stack content */
                for (u = 1; u <= TOS; u++) 
                    fprintf(stdout, "%lld ", Stack[u]);
                    fprintf(stdout, "] -- ");
                    fputc(c, stdout);
            }
            /* debug: print current command to be executed */
            switch (c) {
                /* debug: format debug numbers input */
                case '0'...'9':
                case 'B':
                case 'O':
                    /* ...it's a number prefixed by a letter */
                    if (c1 != ',' && c1 != '@') {    
                        char *p=(input+1);
                        /* (hex number) */
                        if (is_x(*p) && c=='0') {
                            fputc(*p++,stdout);
                            while (isdigit(*p) || ishex(*p)) 
                                fputc(*p++,stdout);
                        }
                        /* (binary, octal or decimal) */
                        else while (isdigit(*p)) fputc(*p++,stdout);
                    }
                    /* ...it's a storing fetching over B,O*/
                    else fputc(c1, stdout);
                    break;
                /* debug: possible >> right shift */
                case '>':
                    if (c1 == c) fputc('>', stdout);
                    break;
                /* debug: possible << left shift */
                case '<':
                    if (c1 == c) fputc('<', stdout);
                    break;
                /* debug: possible #, or #@ commands */
                case '#':
                    if (c1 == ',' || c1 == '@') fputc(c1, stdout);
                    break;
                /* debug: possible @@ or @, commands */
                case '@':
                    if (c1 == c) fputc('@', stdout);
                    else if (c1 == ',') fputc(',', stdout);
                    break;
                /* debug: underscore commands */
                case '_':
                    {
                        /* ... inclusions */ 
                        if (c1 == '[' || c1 == ']') {
                            char *p=input+2;
                            fputc(c1, stdout);
                            while (*p && *p!='[' && *p!= ']') 
                                fputc(*p++, stdout);
                            if (c1 == ']')
                                fprintf(stdout,"[ (inclusion)");
                            else if (c1 == '[') 
                                fprintf(stdout,"] (module)");
                        }
                        /* ... _t time functions */
                        else if (c1 == 't' && strchr("dhmnsyM",*(input+2))) {
                            fputc(c1,stdout);
                            fputc(*(input+2),stdout);
                        }
                        else if (c1 == 't') fputc(c1,stdout);
                        /* ... the _OS command */
                        else if (c1 == 'O' && *(input+2)=='S')
                            fprintf(stdout, "OS");
                        /* ... the _D command */
                        else if (c1 == 'D') printf("D  (debug mode disabled)");
                        /* ... the _r command */
                        else if (c1 == 'r') {
                            if (ROUND == 0.0) 
                                printf("r  (nearest integer rounding set)");
                            else printf("r  (cut rounding enabled set)");
                        }
                        /* ... the _i command */
                        else if (c1 == 'i') {
                            if (isNTDiv) 
                                printf("i  (normal division set)");
                            else printf("i  (integer division set)");
                        }
                        /* ... the _& command */
                        else if (c1 == '&') {
                            if (isPrependAmp) 
                                printf("&  (& before hex disabled)");
                            else printf("&  (& before hex enabled)");
                        }
                        /* ... all others */
                        else if (strchr(",@\'APvcbdohxseq",c1))
                            fputc(c1, stdout);
                        /* ...if not an underscore command, print no op */
                        else fprintf(stdout, " (no op)");
                    }
                    break;
                /* debug: format function loading */
                case '[':
                    fprintf(stdout, " (loading function)");
                    break;
                /* debug: format choice (if) */
                case '\?':
                    {
                        if (c1 == '!') 
                            fputc('!',stdout);
                        else if (BackFuncBuffer[0]!='\0') {
                            if (!Stack[TOS]) 
                                fprintf(stdout," -- [%s]", FuncBuffer);
                            else 
                                fprintf(stdout," -- [%s]", BackFuncBuffer);
                        }
                        else {
                            if(Stack[TOS]) 
                                 fprintf(stdout," -- [%s]", FuncBuffer);
                        }
                    }
                    break;
                /* debug: format cycle (repeat and while) */
                case '!': 
                    {
                        if (c1 == '\?') fputc('\?',stdout);
                        else if (BackFuncBuffer[0]!='\0') 
                            fprintf(stdout, " -- [%s][%s]",\
                                    BackFuncBuffer, FuncBuffer);
                        else fprintf(stdout, " -- [%s]", FuncBuffer);
                    }
                    break;
                /* debug: format strings
                 * warning: only one-line strings will be printed in full;
                 * if a string continues on next line a) the rest is not
                 * printed b) an ellipsis is printed (unfinished string) */ 
                case '\"':
                    {
                        char *p=input+1;
                        while (*p && *p!='\"' && *p!='\n') {
                            if (*p=='\\' && *p=='\"') {
                                fprintf(stdout,"\\\"");
                                p+=2;
                            }
                            else fputc(*p++,stdout);
                        }
                        /* check if double double quotes is present */
                        if (!strncmp(p,"\"\"",2)) fprintf(stdout,"\"\"");
                        /* string completed */
                        else if (*p=='\"') fputc(*p,stdout);
                        /* unfinished string (continuing on next line) */
                        else if (*p=='\n') fprintf(stdout,"...\"");
                    }
                    break;
                /* debug: format storing/fetching 
                 * Note: B and O have already been treated */
                case 'A':
                case 'C'...'N':
                case 'P'...'Z':
                case 'a'...'z':
                    {
                        int c2=*(input+2);
                        /* debug: format storing */
                        if (c1 == ',' && c2 != ',') 
                            fputc(c1, stdout);
                        /* debug: format possible ,, _, _' commands */
                        else if (islow(c) && ((c1 == ',' && c2 == ',') ||\
                                 (c1 == '_' && c2 == ',') ||\
                                 (c1 == '_' && c2 == '\''))) {
                                fputc(c1, stdout);
                                fputc(c2, stdout);
                        }
                    }
                    break;
            }
            /* debug: print variables content if required */
            if (printVars) printvars();
            fputc('\n', stdout);
            dposex=ftell(stdout);
    }
}


/* parse()
 * the parser engine; I analyze each line as a whole entity, for easy managing 
 * of parsing lines and to make the -p option and include easier to implement;
 * return EXIT_SUCCESS if everything went OK; if something went wrong, this 
 * may only due to underflow or overflow errors treated into pop() and push(),
 * so the exit procedure into parse() is only destined to the abort functions
 * ?! and !?, which call in turn quit(); in this case, the timing feature 
 * cannot be reached */
int parse(char *input) {
    /* all-purposes stack-values vars */
    long long int op1, op2, op3;

    /* if we get here, the file is not void! */
    hasExecuted = TRUE;

    /* parse char by char */
    while (*input) {

#ifdef OWL_DEBUG
        /* the debug function if required */
        if (isDebug && !isFunc && !isString && !isComment) debug(input);
#endif

        /* (undocumented)*/
        if (!EXEC) return EXIT_SUCCESS;

        /* The number parser
         * Parse number in either format (binary, octal, decimal, hex)
         * all valid strings are converted to 64 bits numbers */

        /* get a number
         * get a binary number and push it */
        if ('B' == *input && isbin(*(input+1))\
                && !(isString + isComment + isFunc)) {
            char numb[LIMIT], *num = numb;
            input++;
            while (isbin(*input)) *num++ = *input++; 
            *num = '\0';
            push(strtoll(numb, NULL, BIN));
            continue;
        }
        /* get an octal number and push it */
        else if ('O' == *input && isoct(*(input+1))\
                && !(isString + isComment + isFunc)) {
            char numb[LIMIT], *num = numb;
            input++;
            while (isoct(*input)) *num++ = *input++; 
            *num = '\0';
            push(strtoll(numb, NULL, OCT));
            continue;
        }
        /* get a hexadecimal number and push it */
        else if ('0' == *input && is_x(*(input+1))\
                   && !(isString + isComment + isFunc)) {
            char numb[LIMIT], *num = numb;
            input += 2;
            while ((isdigit(*input)) || ishex(*input)) *num++ = *input++; 
            *num = '\0';
            push(strtoll(numb, NULL, HEX));
            continue;
        }
        /* get a decimal number and push it */
        else if (isdigit(*input) && !(isString + isComment + isFunc)) {
            char numb[LIMIT], *num = numb;
            while ((isdigit(*input)) || (*input=='E')) *num++ = *input++; 
            *num = '\0';
            push(strtoll(numb, NULL, DEC));
            continue;
        }
        
        /* the functions parser
         * get a function and store it into the function buffer; 
         * skip possible inclusion commands into function */

        /* init: initialize the function buffer, copying its content
         * to the back-buffer, resetting function buffer and pointer */
        if (('[' == *input) && (!isFunc) && !(isComment + isString)) {
            int i;
            FBuffer = FuncBuffer;
            BFBuffer = BackFuncBuffer;
            for (i = 0; i < LIMIT; i++) *BFBuffer++ = *FBuffer++;
            isFunc = 1;
            erase_buffer(FuncBuffer);
            FBuffer = FuncBuffer;
            input++;
            continue;
        }
        /* check end of function */
        else if ((1 == isFunc) && !(isComment + isString) && !isInclude &&\
                ((']' == *input) ||\
                ((FBuffer - FuncBuffer) >= LIMIT - 1))) {
            *FBuffer = '\0';
            isFunc = 0;
            if (*input==']') input++;
            continue;
        }
        /* store function */
        else if (isFunc && !(isComment + isString)) {
            if ('[' == *input && !isInclude) isFunc++;
            else if ('[' == *input &&  isInclude) isInclude=FALSE;
            else if (']' == *input && !isInclude) isFunc--;
            else if (']' == *input &&  isInclude) isInclude=FALSE;
            /* skip inclusion command */
            else 
                if (('_'==*input) && ('['==*(input+1) || ']'==*(input+1))){
                isInclude=TRUE;
                *FBuffer++ = *input++;
            }
            *FBuffer++ = *input++;
            continue;
        } 

        /* the string parser
         * get a string and store it into PAD;
         * parse also all implemented escape chars */

        /* init: initialize erasing PAD (if required) */
        if ((!isString) && ('\"' == *input) && !(isComment + isFunc)) {
            isString = TRUE;
            /* use -e option at start for erasing strings before 
             * instantiating; this prevents the use of PAD as an array,
             * because any string would blank the array when instantiated */
            if (isBlank) erase_buffer(PAD);
            Pad = PAD;
            input++;
            continue;
        }
        /* check end of string */
        else if (isString && !(isComment + isFunc) &&\
                (('\"' == *input) ||\
                ((Pad - PAD) >= LIMIT - 1))) {
            *Pad = '\0';
            if ('\"' == *(input+1)) input++;
            else {
                /* a double couple "" (the void string) acts as "\n" */
                if (!strlen(PAD)) fprintf(stdout, "\n");
                else fprintf(stdout, "%s", PAD);
                //fprintf(stdout, "%s", PAD);
                fflush(NULL);
            }
            isString = FALSE;
            if (*input=='\"') input++;
            continue;
        }
        /* store string */
        else if (isString && !(isComment+isFunc)) {
            /* parse escape chars */
            if ('\\' == *input) {
                switch(*(input + 1)) {
            case '0': *Pad++ = '\0'; break; // placeholder
                    case 'b': *Pad++ = '\b'; break; // backspace
                    case 'f': *Pad++ = '\f'; break; // form feed
                    case 'n': *Pad++ = '\n'; break; // newline
                    case 'r': *Pad++ = '\r'; break; // carriage return
                    case 't': *Pad++ = '\t'; break; // horizontal tab
                    case 'v': *Pad++ = '\v'; break; // vertical tab
                    case '\"':*Pad++ = '\"'; break; // double quote
                    case '\'':*Pad++ = '\''; break; // single quote
                    case '\\':*Pad++ = '\\'; break; // backslash
                    case '\?':*Pad++ = '\?'; break; // question mark
                    /* extended escape chars according 
                     * to a common extended table */
                    case 'c': *Pad++ = 162; break;
                    case 'L': *Pad++ = 163; break;
                    case 'S': *Pad++ = 167; break;
                    case '<': *Pad++ = 171; break;
                    case 'm': *Pad++ = 172; break;
                    case 's': *Pad++ = 175; break;
                    case 'o': *Pad++ = 176; break;
                    case '+': *Pad++ = 177; break;
                    case '2': *Pad++ = 178; break;
                    case '3': *Pad++ = 179; break;
                    case 'u': *Pad++ = 181; break;
                    case 'x': *Pad++ = 183; break;
                    case '>': *Pad++ = 187; break;
                    case 'A': *Pad++ = 198; break;
                    case 'p': *Pad++ = 215; break;
                    case 'B': *Pad++ = 223; break;
                    case 'a': *Pad++ = 230; break;
                    case '-': *Pad++ = 247; break;
                    case 'O': *Pad++ = 248; break;
                    case 'T': *Pad++ = 254; break;
                    default:  
                        *Pad++ = '\\'; 
                        *Pad++ = *(input + 1);
                }
                input++;
            }
            else *Pad++ = *input;
            input++;
            continue;
        }
        
        /* multiline comment parsing 
         * parse all characters until a closing token found, doing nothing 
         * in between; Note: the single line comment is detected 
         * into the language process routine */

        /* check where comment begins */
        if ('(' == *input && '*' == *(input+1) && !isComment\
                && !(isString + isFunc)) {
            isComment = TRUE;
            input += 2;
            continue;
        }
        /* check where comment ends */
        else if ('*' == *input && ')' == *(input+1) && isComment\
                && !(isString + isFunc)) {
            isComment = FALSE;
            input += 2;
            continue;
        }
        /* discard inner chars */
        else if (isComment && !(isString + isFunc)) {
            input++;
            continue;
        } 
        
        /* language process routine
         * parse char by char and execute code
         * cases are in ASCII order */
        switch (*input) {
            /* for now, we discard any blank or control char */
            case 1 ... 32: break;
            case '!': /* cycles and abort */
                {
                    char fun[LIMIT], bfun[LIMIT];
                    /* !? abort */
                    if ('?' == *(input+1)) { quit(); }
                    /* cycles */
                    strncpy(fun, FuncBuffer, LIMIT);
                    strncpy(bfun, BackFuncBuffer, LIMIT); 
                    erase_buffer(FuncBuffer);
                    erase_buffer(BackFuncBuffer);
                    /* []! begin until */
                    if ('\0' == bfun[0]) {
                        do {
                            parse(fun);
                            op1 = pop();
                        }
                        while (!op1);
                    }
                    /* [][]! begin while repeat */
                    else {
                        parse(bfun);
                        op1 = pop();
                        while (op1) {
                            parse(fun);
                            parse(bfun);
                            op1 = pop();
                        }
                    }
                }
                break;
            /* $ swap */
            case '$': 
                op1 = pop();
                op2 = pop();
                push(op1);
                push(op2);
                break;
            /* % dup */
            case '%': 
                op1 = pop();
                push(op1);
                push(op1);
                break;
            /* & bitwise and */
            case '&': 
                op1 = pop();
                op2 = pop();
                push(op1 & op2);
                break;
            /* ' roll */ 
            case '\'': 
                op2 = pop();
                if (TOS-op2>0) op3 = Stack[TOS - op2];
                else break;
                for (op1 = 1; op1 <= op2; op1++)
                    Stack[TOS - op2 + op1 - 1] = Stack[TOS - op2 + op1];
                Stack[TOS] = op3;
                break;
            /* ( input char - one char only */
            case '(': 
#ifndef OWL_REMOVE
                {
                    // thanks to Kyle R. Burton
                    // for this routine. See initial
                    // GPL banner for credits
                    struct termios modes, savemodes;
                    tcgetattr(0, &modes);
                    savemodes = modes;
                    modes.c_lflag &= ~ICANON;
                    modes.c_lflag &= ~ECHO;
                    tcsetattr(0, 0, &modes);
#endif
                    op1 = fgetc(stdin);
                    push(op1);
#ifndef OWL_REMOVE
                    tcsetattr(0, 0, &savemodes);
                }
#endif
                break;
            /* ) emit as char */
            case ')': 
                op1 = pop();
                /* this piece of code makes any CR/NL output portable on 
                 * any Operating System. In ASCII, 13 is CR, 10 is NL */
                if (13 == op1 || 10 == op1) fputc('\n', stdout);
                else fputc((char)op1, stdout);
                fflush(NULL);
                break;
            /* * multiplication */
            case '*': 
                op1 = pop();
                op2 = pop();
                push(op1 * op2);
                break;
            /* + addition */
            case '+': 
                op1 = pop();
                op2 = pop();
                push(op1 + op2);
                break;
            case ',': 
                /* store a value into PAD */
                op1 = pop();
                op2 = pop();
                PAD[(int)llabs(op1)%LIMIT] = (char)op2;
                break;
            /* - subtraction */
            case '-': 
                op1 = pop();
                op2 = pop();
                push(op2 - op1);
                break;
            case '.': 
                /* . print number (according to base) */
                op1 = pop();
                switch (base) {
                    case DEC:
                        fprintf(stdout, "%lld", op1);
                        break;
                    case BIN:
                        fprintb(stdout, op1);
                        break;
                    case OCT:
                        fprintf(stdout, "%llo", op1);
                        break;
                    case HEX:
                        /* I prefer capital letters in hex numbers */
                        if (isPrependAmp)
                            fprintf(stdout, "&%llX", op1);
                        else 
                            fprintf(stdout, "%llX", op1);
                        break;
                }
                fflush(NULL);
                break;
            /* / division */
            case '/': 
                op1 = pop();
                op2 = pop();
                push(divide(op2,op1));
                break;
            /* : root */
            case ':': 
                op1 = pop();
                op2 = pop();
                push(nroot(op2,op1));
                break;
            /* ; drop */
            case ';': 
                pop();
                break;
            case '<':
                /* << shift left */
                if (*(input+1) == '<') {
                    input++;
                    op1 = pop();
                    op2 = pop();
                    push(op2 << op1);
                }
                /* < input a number */
                else {
                    char NUM[LIMIT];
                    fgets(NUM, LIMIT, stdin);
                    if ('B' == NUM[0] && isbin(NUM[1]))
                        push(strtoll(NUM+1, NULL, BIN));
                    else if ('O' == NUM[0] && isoct(NUM[1]))
                        push(strtoll(NUM+1, NULL, OCT));
                    else if ('0' == NUM[0] && is_x(NUM[1]))
                        push(strtoll(NUM+2, NULL, HEX));
                    else push(strtoll(NUM, NULL, DEC));
                }
                break;
            /* = equality test */
            case '=': 
                op1 = pop();
                op2 = pop();
                push(-(op1==op2));
                break;
            case '>': 
                op1 = pop();
                op2 = pop();
                /* >> shift right */
                if (*(input+1) == '>') {
                    input++;
                    op3 = op2 >> op1;
                }
                /* > greater than test */
                else {
                    op3 = -(op2 > op1);
                }
                push(op3);
                break;
            /* choices and abort */
            case '\?': 
                {
                    char fun[LIMIT], bfun[LIMIT];
                    /* ?! abort */
                    if ('!' == *(input+1)) { quit(); }
                    /* choices */
                    strncpy(fun, FuncBuffer, LIMIT);
                    strncpy(bfun,BackFuncBuffer, LIMIT); 
                    erase_buffer(FuncBuffer);
                    erase_buffer(BackFuncBuffer);
                    op1 = pop();
                    /* []? if then */
                    if ('\0' == bfun[0]) {
                        if (op1) parse(fun);
                    }
                    /* [][]? if else then */
                    else {
                        if (op1) parse(bfun);
                        else parse(fun);
                    }
                }
                break;
            case '@': 
                /* @@ dynamic binding - execute indexed function */
                if ('@' == *(input+1)) {
                    parse(VarFBuffer[fIndex - 'a']);
                    input++;
                }
                /* fetch a value from PAD */ 
                else {
                    op1 = pop();
                    push((long long int)PAD[(int)llabs(op1)%LIMIT]);
                }
                break;
            case '#':
                /* store a value into integer array */
                if (*(input+1)==',') {
                    op1 = pop();
                    op2 = pop();
                    PADARR[(int)llabs(op1)%ARRAYLIM] = op2;
                    input++;
                }
                /* fetch a value from integer array */ 
                else if (*(input+1)=='@') {
                    op1 = pop();
                    push(PADARR[(int)llabs(op1)%ARRAYLIM]);        
                    input++;
                }
                /* # single line comment: break immediately */
                else return EXIT_SUCCESS;
                break;
            /* A..Z numeric variables */
            case 'A' ... 'Z': 
                input = getVar(input);
                break;
            /* \ negate */
            case '\\': 
                op1 = pop();
                push(-op1);
                break;
            /* ^ power */
            case '^': 
                op1 = pop();
                op2 = pop();
                push(power(op2, op1));
                break;
            case '_':
                /* _ underscore (system) commands 
                 * (or commands also beginning with underscore) */
                switch (*(++input)) {
                    case ']':
                        /* include file whose name is in the form _]inc.owl[
                         * in the stream, without a copy of local variables */
                        input += include_file(input + 1) + 1;
                        break;
                    case '[':
                        /* execute file whose name is in the form _[inc.owl] 
                         * in the stream, making a copy of local variables */
                        {
                            int i;
                            int  TempIBuffer[MAXVARS];
                            char TempFBuffer[MAXVARS][LIMIT];
                            /* copy caller vars into temp buffers */
                            for (i = 0; i < MAXVARS; i++) {
                                TempIBuffer[i] = VarIBuffer[i];
                                strncpy(TempFBuffer[i], VarFBuffer[i], LIMIT);
                            }
                            /* execute file as usual with a copy of vars */
                            input += include_file(input + 1) + 1;
                            /* restore caller vars */
                            for (i = 0; i < MAXVARS; i++) {
                                VarIBuffer[i] = TempIBuffer[i];
                                strncpy(VarFBuffer[i], TempFBuffer[i], LIMIT);
                            }
                        }
                        break;
#ifndef OWL_REMOVE
                    /* _c chronometer command */
                    case 'c':
                         /* it's the odd invocation (reset chronometer) */
                        if (chrono & 1) {
                            gettimeofday(&tvChrono, NULL);
                            begChrono = ((double)tvChrono.tv_sec\
                                    + tvChrono.tv_usec / 1000000.0);
                          }
                         /* it's the even invocation (return milliseconds) */
                        else {
                            gettimeofday(&tvChrono, NULL);
                            endChrono = ((double)tvChrono.tv_sec\
                                    + tvChrono.tv_usec / 1000000.0);
                            push((int)((endChrono - begChrono) * 1000));
                          }
                        chrono++;
                        break;
#endif
                    /* _@ dynamic binding: execute PAD as function */
                    case '@':
                        parse(PAD);
                        break;
                    /* _A push array upper limit */
                    case 'A':
                        push((long long int)ARRAYLIM);
                        break;
                    /* _P push PAD upper limit */
                    case 'P':
                        push((long long int)LIMIT);
                        break;
                    /* _D toggle debug mode */
                    case 'D':
                        isDebug = (isDebug==TRUE?FALSE:TRUE);
                        break;
                    case 'O':
                        /* _OS push OS type 0 = UNIX ; 1 = WIN */
                        if ('S' == *(input + 1)) {
#ifndef OWL_REMOVE
                            push(0ll);
#else
                            push(1ll);
#endif
                            ++input;
                        }
                        break;
                    /* _v push version numbers */
                    case 'v':
                        push((long long int)RELEASE);
                        push((long long int)MINOR);
                        push((long long int)MAJOR);
                        break;
                    /* _b set binary base for output */
                    case 'b':
                        base = BIN;
                        break;
                    /* _d set base decimal for output */
                    case 'd':
                        base = DEC; 
                        break;
                    /* _o set octal base for output */
                    case 'o':
                        base = OCT; 
                        break;
                    /* _h _x set hex base for output */
                    case 'h': 
                    case 'x': 
                        base = HEX; 
                        break;
                    /* _s execute PAD as shell command */
                    case 's':
                        system(PAD);
                        break;
                    /* _e clear both PAD and array */
                    case 'e':
                        erase_buffer(PAD);
                        for(op1=0; op1<ARRAYLIM; op1++) PADARR[op1]=0LL;
                        break;
                    /* _i toggle integer division mode */
                    case 'i':
                        isNTDiv = (isNTDiv ? FALSE : TRUE); 
                        break;
                    /* _r toggle rounding mode */
                    case 'r':
                        ROUND = (0.0L == ROUND ? 0.5L : 0.0L);
                        break;
                    /* _& toggle prepending & in hex numbers output */
                    case '&':
                        isPrependAmp = (isPrependAmp ? FALSE : TRUE);
                        break;
                    /* _q return stack depth */
                    case 'q':
                        push(TOS);
                        break;
#ifndef OWL_REMOVE
                    /* time values */
                    case 't':  
                        {
                            struct tm *st_ptr;
                            time_t st_lt;
                            long long int hour, min, sec, msec;;
                            long long int day, month, year;
                            struct timeval tv;
                        
                            st_lt  = time(NULL);
                            st_ptr = localtime(&st_lt); 
                            gettimeofday(&tv, NULL);
                            hour = st_ptr->tm_hour;
                            min  = st_ptr->tm_min;
                            sec  = st_ptr->tm_sec;
                            msec = tv.tv_usec;
                            year =st_ptr->tm_year + 1900;
                            month= st_ptr->tm_mon + 1;
                            day  = st_ptr->tm_mday;
                            /* _tn push microseconds */
                            if ('n' == *(input+1)) {
                                push(msec);
                                input++;
                            }
                            /* _th push hour */
                            else if ('h' == *(input+1)){
                                push(hour);
                                input++;
                            }
                            /* _tm push minutes */
                            else if ('m' == *(input+1)){
                                push(min);
                                input++;
                            }
                            /* _ts push seconds */
                            else if ('s' == *(input+1)){
                                push(sec);
                                input++;
                            }
                            /* _td push day */
                            else if ('d' == *(input+1)){
                                push(day);
                                input++;
                            }
                            /* _tM push month */
                            else if ('M' == *(input+1)){
                                push(month);
                                input++;
                            }
                            /* _ty push year */
                            else if ('y' == *(input+1)){
                                push(year);
                                input++;
                            }
                            /* _t push year,month,day,sec,min,hour */
                            else {
                                push(year);
                                push(month);
                                push(day);
                                push(sec);
                                push(min);
                                push(hour);
                            }
                        }
                        break;
#endif
                    /* ignore all the rest */
                    default:
                        input--;
                        break;
                }
                break;
            /* ` pick */
            case '`': 
                op1 = pop();
                if (TOS-op1>0) push(Stack[TOS - op1]);
                break;
            /* a..z function variables */
            case 'a' ... 'z': 
                input = getVar(input);
                break;
            /* { input a string
               also get rid of terminating return key left by fgets() */
            case '{': 
                fgets(PAD, LIMIT, stdin);
                PAD[strlen(PAD) - 1] = '\0';
                break;
            /* | bitwise or */
            case '|': 
                op1 = pop();
                op2 = pop();
                push(op1 | op2);
                break;
            /* } print string in PAD
               no control is done to check if string is terminated by '\0' */
            case '}': 
                fprintf(stdout, "%s", PAD);
                fflush(NULL);
                break;
            /* ~ logical not */
            case '~': 
                op1 = pop();
                push(-(!op1));
                break;
        }
        input++;
    }
   

    return EXIT_SUCCESS;
}


/*************************
 *        MAIN           *
 *************************/ 
/* hic sunt leones */
int main(int argc, char **argv) {
    
    FILE *f = NULL;
    int c, ac = argc;
    char line[LIMIT];
#ifndef OWL_REMOVE
    struct timeval tv;           /* time values        */
    double beginS, endS;         /* reports in seconds */
    boolean isTiming = FALSE;    /* timer flag         */
#endif

    /* initialize all variables, reset stack, initialize arrays and
     * erase functions/strings buffers */
    for (c = 0; c <= MAXVARS; c++) {
        VarIBuffer[c] = 0LL;
        erase_buffer(VarFBuffer[c]);
    }
    for(c = 0; c < LIMIT; c++) Stack[c] = 0LL;  
    for(c = 0; c<ARRAYLIM; c++) PADARR[c]=0LL;
    erase_buffer(BackFuncBuffer);
    erase_buffer(FuncBuffer);
    erase_buffer(PAD);
    
    /* parse interpreter arguments */
    while (--ac > 0 && '-' == (*++argv)[0]) {
        /* parse double-dash options */
        if (!strcmp(*argv, "--help")) {
            printHelp();
            exit(EXIT_SUCCESS);    
        }
        else if (!strcmp(*argv, "--version")) {
            printVersion(1);
            exit(EXIT_SUCCESS);    
        }
        else if (!strcmp(*argv, "--conditions")) {
            printConditions(1);
            exit(EXIT_SUCCESS);    
        }
        else if (!strcmp(*argv, "--warranty")) {
            printWarranty(1);
            exit(EXIT_SUCCESS);    
        }
        /* this feature is intentionally undocumented 
         * I use it to generate the README file */
        else if (!strcmp(*argv, "--greetings")) {
            printf("Welcome, owl user! I hope you\'ll like ");
            printf("this program.\n");
            printf("Here\'s some information about it ");
            printf("and its license...\n");
            printVersion(0);
            printHelp();
            printf("To compile type 'make' as user and as ");
            printf("root 'make install' in the owl directory.\n");
            printf("For help about owl's features, after");
            printf(" installation, see the manual.\n\n");
            printWarranty(0);
            printConditions(0);
            printf("It's GPL. Enjoy!\n");
            exit(EXIT_SUCCESS);
        }
        
        /* parse single-dash options */ 
        while ((c = *++argv[0])) {
            switch(c) {
#ifdef LEAVESEXE
                case 'c': leavesEXE = TRUE; break;
#endif
#ifdef OWL_DEBUG
                case 'd': isDebug = TRUE; break;
                case 'D': printVars = TRUE; break;
#endif
                case 'e': isBlank = (isBlank ? FALSE : TRUE); break;
                case 'h': printHelp(); exit(EXIT_SUCCESS);
                /* toggle integer division type */
                case 'i': isNTDiv = (isNTDiv ? FALSE : TRUE); break;
                /* parse string and exit */
                case 'p':
#ifndef OWL_REMOVE
                    if (isTiming) {
                        gettimeofday(&tv, NULL);
                        beginS=((double)tv.tv_sec+tv.tv_usec/1000000.0);
                    }
#endif
                    if ('\0' == (*argv)[1]) {
                        if ('\0'==*(argv+1)) {
                            fprintf(stderr, NOSTRING);
                            fprintf(stderr, USEHELP);
                            exit(EXIT_FAILURE);
                        }
                        else parse(*(argv+1));
                    }
                    else parse(*argv+1);
#ifndef OWL_REMOVE
                    if (isTiming) goto pexit;
                    else 
#endif                            
                    { 
                        fprintf(stdout, "\n");
                        exit(EXIT_SUCCESS);
                    }
                /* toggle rounding feature */
                case 'r': ROUND = (0.0L == ROUND ? 0.5L : 0.0L); break;
#ifdef LEAVESASM
                case 's': leavesASM = TRUE; break;
#endif
#ifndef OWL_REMOVE
                case 'S': singleStep = TRUE; break;
                case 't': isTiming = TRUE; break;
#endif
                case 'v': printVersion(1); exit(EXIT_SUCCESS);
                /* Option -X is undocumented: if used, the debugger 
         * is enabled and parse() exits after launching 
         * the debugger on current line */
                case 'X': EXEC = FALSE; isDebug = TRUE; break;
                default:
                    fprintf(stderr, WRONGOPTERROR, c);
                    fprintf(stderr, USEHELP);
                    exit(EXIT_FAILURE);
            }
        }
    }

    /* get input file name (discard following names) */ 
    FileName = *argv;

    /* OK; now, if after any option there's no file name left
     * (or if owl has been launched with no parameters), issue an error */
    if (('\0' == *FileName) || (1 == argc)) {
        fprintf(stderr, NOFILE);
        fprintf(stderr, USEHELP);
        exit(EXIT_FAILURE);
    }


    if (FileName) {
        f = fileopen(FileName);
        if (!f) {    
            fprintf(stderr, OPENFILEERROR, FileName);
            fprintf(stderr, USEHELP); 
            exit(EXIT_FAILURE);
        }
    }

    /* data (numbers and instructions) after file name, are passed to 
     * owl programs as parameters; particularly useful for scripts 
     * (provided you escape characters having some meaning for the shell) */
    for (argc = 1; argc < ac; argc++) parse(*(argv + argc));

#ifndef OWL_REMOVE
    /* record start timer data if required */
    if (isTiming) {
        gettimeofday(&tv, NULL);
        beginS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);
    }
#endif
    
    /* process file lines and close after EOF */
    /* This is the heart of the process, and see how slim and terse it is! */
    while (fgets(line, LIMIT, f) && !parse(line));
    fclose(f);
    
#ifndef OWL_REMOVE
pexit:
    /* record end timer data if required, and print time report */
    if (isTiming) {
        int ptime;
        gettimeofday(&tv, NULL);
        endS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);
        ptime = (int)((endS - beginS) * 1000);
        if (ptime < 1000) {
            /* time reported in milliseconds... */
            fprintf(stdout, TIMES, ptime);
        }
        else {
            /* ...or time reported in seconds */
            fprintf(stdout, TIMESS, ((double)ptime / 1000));
        }
    }
#endif

    /* print eventual unfinished string in the last line of owl code */
    if (isString) fprintf(stdout, "%s\n", PAD);
   
    /* print a carriage return, plus an advice 
     * if no char was processed, and exit */
    if (!hasExecuted) fprintf(stdout, "(no code executed)\n");    
    else fprintf(stdout,"\n");
    return EXIT_SUCCESS;

} 
/* end main */
